/*
 * $Header: /Book/AbstractEditWnd.cpp 20    11.07.01 14:26 Oslph312 $
 * Contains functionality common to edit and rich edit controls.
 * TODO: Split into a _really_ abstract base class and a concrete impl for Edit & RE
 */

#include "precomp.h"
#include "AbstractEditWnd.h"
#include "EditWnd.h"
#include "RichEditWnd.h"
#include "ClientDC.h"
#include "InstanceSubclasser.h"
#include "search.h"


#ifdef _DEBUG
bool AbstractEditWnd::bForceEdit = false;
#endif


PRIVATE LRESULT CALLBACK editParentSpy( 
   HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam );

PRIVATE InstanceSubclasser parentSpy( editParentSpy );

PRIVATE LRESULT CALLBACK editParentSpy( 
   HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam ) 
{
   assert( IsWindow( hwnd ) );
   if ( WM_COMMAND == msg ) {
      AbstractEditWnd *pEditWnd = reinterpret_cast< 
         AbstractEditWnd * >( parentSpy.getUserData( hwnd ) );
      if ( 0 != pEditWnd ) {
         assert( isGoodPtr( pEditWnd ) );

         const HWND hwndCtl = GET_WM_COMMAND_HWND( wParam, lParam );
         const int  idCtl   = GET_WM_COMMAND_ID  ( wParam, lParam );
         if ( (HWND) *pEditWnd == hwndCtl &&
              GetWindowID( *pEditWnd ) == idCtl )
         {
            assert( IsWindow( *pEditWnd ) );
            assert( GetParent( *pEditWnd ) == hwnd );
            EditListener *pListener = pEditWnd->getEditListener();
            assert( isGoodPtr( pListener ) );

            switch ( GET_WM_COMMAND_CMD( wParam, lParam ) ) {
            case EN_CHANGE  : return pListener->onChange  (), 0;
            case EN_ERRSPACE: return pListener->onErrSpace(), 0;
            case EN_MAXTEXT : return pListener->onMaxText (), 0;
            default:
               break;
            }
         }
      }
   }
   return parentSpy.callOldProc( hwnd, msg, wParam, lParam );
}


/**
 * Static factory method to create 
 * a concrete implementation of an AbstractEditWnd.
 */
AbstractEditWnd *AbstractEditWnd::create( 
   HWND hwndParent, LPCTSTR pszText, 
   EditListener *pEditListener, bool bWordWrap ) 
{
   const HMODULE hmodule = 

#ifdef _DEBUG
      bForceEdit ? 0 :
#endif

      GetModuleHandle( _T( "RICHED20.DLL" ) );

   AbstractEditWnd *pEditWnd = 0;
   if ( 0 != hmodule ) {
      pEditWnd = createRichEdit( hwndParent, pszText );
   } else {
      pEditWnd = createEdit( hwndParent, pszText, bWordWrap );
   }
   assert( 0 != pEditWnd );
   assert( 0 != pEditListener );
   pEditWnd->m_pEditListener = pEditListener;
   parentSpy.subclass( hwndParent, pEditWnd );
   return pEditWnd;
}


/**
 * The whole point of this method is to be able to update the 
 * line and column numbers on the status bar, and to enable or
 * disable the cut and copy commands on the tool bar.
 *
 * We don't get any notifications from the edit control when the 
 * user moves the cursor, for example, so we just listen to all 
 * messages and  check if the current position has changed.
 */
LRESULT AbstractEditWnd::dispatch( 
   UINT msg, WPARAM wParam, LPARAM lParam ) 
{
   assertValid();
   static bool bLock = false;
   if ( !bLock ) {
      bLock = true;

      int nStart = 0;
      int nEnd = 0;
      getSel( &nStart, &nEnd );

      const int nLine = lineFromChar( nStart );
      const int nColumn = nStart - sendMessage( EM_LINEINDEX, nLine );
      assert( 0 <= nLine );
      assert( 0 <= nColumn );
      if ( nColumn  != m_nCurColumn || 
           nLine    != m_nCurLine   || 
           nEnd     != m_nLastEnd   ) 
      {
         EditListener *pListener = getEditListener();
         if ( 0 != pListener ) { // May be 0 during startup
            m_nCurLine   = nLine;
            m_nCurColumn = nColumn;
            m_nLastEnd   = nEnd;
            pListener->onPosChange( Point( nColumn, nLine ) );
         } else {
            resetPosition();
         }
      }
      bLock = false;
   }

   return Window::dispatch( msg, wParam, lParam );
}


void AbstractEditWnd::scroll( UINT nWhat ) { 
   
   assertValid();
   sendMessage( WM_VSCROLL, 
      MAKEWPARAM( nWhat, 0 ), (LPARAM)(HWND)*this );
}


int AbstractEditWnd::getVisibleLineCount( HFONT hfont ) const {

   assertValid();
   assert( 0 != hfont );
   
   int nVisibleLineCount = 0;
   RECT rc;
   sendMessage( EM_GETRECT, 0, reinterpret_cast< LPARAM >( &rc ) );
   const int nHeight = rc.bottom - rc.top;

   ClientDC dc( *this );
   if ( dc.isValid() ) {
      HFONT hfontSaved = SelectFont( dc, hfont );
      TEXTMETRIC textMetric;
      bool bOK = 0 != GetTextMetrics( dc, &textMetric );
      SelectFont( dc, hfontSaved );
   
      if ( bOK ) {
         const int nLineHeight = 
            textMetric.tmHeight + textMetric.tmExternalLeading;
         assert( 0 < nLineHeight );
         if ( 0 < nLineHeight ) {
            nVisibleLineCount = nHeight / nLineHeight;
         }
      }
   }

   return nVisibleLineCount;
}


/**
 * Helper method for bringCaretToWindow.
 * Returns new, adjusted insertion point.
 */
int AbstractEditWnd::moveInsertionPoint( int nLine, int nColumn ) {

   assertValid();
   const int nLineStart = sendMessage( EM_LINEINDEX, nLine );
   const int nLineLength = sendMessage( EM_LINELENGTH, nLineStart );
   if ( nLineLength < nColumn ) {
      nColumn = nLineLength;
   }
   return  nLineStart + nColumn;
}


/**
 * Bring the caret into view by moving the caret position,
 * not by scrolling the text in the window, as EM_SCROLLCARET does.
 * Takes no action if there is an actual selection, as it would
 * screw up the selecion.
 */
void AbstractEditWnd::bringCaretToWindow( HFONT hfont ) {

   assertValid();
   int nSelStart, nSelEnd;
   if ( !getSel( &nSelStart, &nSelEnd ) ) {
      assert( nSelStart == nSelEnd );

      const int nTop = sendMessage( EM_GETFIRSTVISIBLELINE );
      const int nCurLine = sendMessage( EM_LINEFROMCHAR, nSelStart );
      int nCurColumn = 
         nSelStart - sendMessage( EM_LINEINDEX, nCurLine );

      if ( nCurLine < nTop ) {
         nSelStart = moveInsertionPoint( nTop, nCurColumn );
         setSel( nSelStart, nSelStart );
      } else {
         const int nLines = getVisibleLineCount( hfont );
         const int nLast = nTop + nLines - 1;
         if ( nLast < nCurLine ) {
            nSelStart = moveInsertionPoint( nLast, nCurColumn );
            setSel( nSelStart, nSelStart );
         }
      }
   }
}


int AbstractEditWnd::getLineCount( void ) const {
   
   assertValid();
   const int nChars = getTextLength();
   if ( 0 == nChars ) {
      return 0;
   }
   return lineFromChar( nChars );
}


int AbstractEditWnd::getTextLength( bool ) const {

   return sendMessage( WM_GETTEXTLENGTH );
}


class EditContents {
private:
   const AbstractEditWnd *m_pEditWnd;
   LPTSTR m_psz;
   int m_nLength;

public:
   EditContents( const AbstractEditWnd *pEditWnd ) 
      : m_pEditWnd( pEditWnd )
      , m_psz( 0 ) 
   {
      lock();
   }
   ~EditContents() {
      release();
   }
   int length( void ) {
      return m_nLength;
   }
   void lock( void ) {
      const int nLength = m_pEditWnd->getTextLength();
      m_psz = new TCHAR[ nLength + 1 ];
      m_nLength = m_pEditWnd->getSearchText( m_psz, nLength + 1 );
      assert( m_nLength <= nLength );
      assert( isGoodStringPtr( m_psz ) );
      assert( 0 == m_psz[ m_nLength ] );
      m_psz[ m_nLength ] = 0;
   }
   void release( void ) {
      assert( isGoodStringPtr( m_psz ) );
      delete[] m_psz;
      reset_pointer( m_psz );
   }
   operator LPCTSTR() {
      assert( isGoodStringPtr( m_psz ) );
      return m_psz;
   }
};


bool AbstractEditWnd::searchAndSelect( 
   const String& strSearchPattern, 
   const bool bMatchWholeWord, 
   const bool bMatchCase,
   const int  nDirection )
{
   if ( 0 != nDirection && 0 == strSearchPattern.length() ) {
      return false;
   }

   // Find the current selection range and text length:
   int nStart = 0;
   const bool bHasSelection = getSel( &nStart );

   EditContents pszText( this );
   const int nLength = pszText.length();

   LPCTSTR pszFound = 0;
   int nFound = nStart;

   const int nSearchLength = strSearchPattern.length();
   assert( 0 < nSearchLength );
   LPCTSTR pszSearchPattern = strSearchPattern.c_str();
   if ( 0 < nDirection ) {   // Forward search.
      LPCTSTR pszStart = pszText + nStart + (bHasSelection ? 1 : 0);
      pszFound = find( pszText, pszStart, pszSearchPattern, 
         bMatchWholeWord, bMatchCase );
      if ( 0 == pszFound && (LPCTSTR) pszText < pszStart ) {
         pszFound = find( pszText, pszText, pszSearchPattern, 
            bMatchWholeWord, bMatchCase );
         assert( 0 == pszFound || pszFound - pszText <= nStart );
      }
   } else {                  // Backwards search
      if ( 0 < nStart ) {
         pszFound = findBackwards( pszText, pszText + nStart - 1, 
            pszSearchPattern, bMatchWholeWord, bMatchCase );
      }
      if ( 0 == pszFound ) {
         pszFound = findBackwards( pszText, pszText + nLength, 
            pszSearchPattern, bMatchWholeWord, bMatchCase );
      }
   }

   if ( 0 != pszFound ) {
      nFound = pszFound - pszText;
      setSel( nFound, nFound + nSearchLength );
      sendMessage( EM_SCROLLCARET );
   }

   return 0 != pszFound;
}


int AbstractEditWnd::replaceInSelection( 
   const String& strSearchPattern, 
   const String& strReplacePattern,
   const bool bMatchWholeWord, 
   const bool bMatchCase )
{
   int nStart = 0;
   int nEnd   = 0;
   verify( getSel( &nStart, &nEnd ) );

   const int nSearchLength = strSearchPattern.length();
   assert( 0 < nSearchLength );
   LPCTSTR pszSearchPattern = strSearchPattern.c_str();

   const int nReplaceLength = strReplacePattern.length();
   LPCTSTR pszReplacePattern = strReplacePattern.c_str();

   const int nExtra = nReplaceLength - nSearchLength;

   EditContents pszText( this );

   for ( int nReplacements = 0; ; ++nReplacements ) {
      const LPCTSTR pszStart = pszText + nStart;
      const LPCTSTR pszFound = find( pszText,
         pszStart, pszSearchPattern, bMatchWholeWord, bMatchCase );
      if ( 0 == pszFound ) {
         break; //*** LOOP EXIT POINT
      }
      const int nFound = pszFound - (LPCTSTR) pszText;
      assert( nStart <= nFound );
      if ( nEnd < nFound + nSearchLength ) {
         break; //*** LOOP EXIT POINT
      }

      pszText.release(); 
      setSel( nFound, nFound + nSearchLength );
      replaceSel( pszReplacePattern );
      pszText.lock();
      nStart = nFound + nReplaceLength;
      nEnd += nExtra;
   }

   return nReplacements;
}

// end of file
